
# Library imports
import os
import pickle
import random
import torch
import string
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
import networkx as nx
import pandas as pd
import os.path as osp
from scipy.stats import rankdata
from sklearn.metrics import roc_auc_score, average_precision_score
from torch_geometric.datasets import Planetoid, WordNet18RR, RelLinkPredDataset
import torch_geometric.transforms as T
from torch_geometric.nn import GCNConv
from torch_geometric.data import Data
from torch_geometric.utils import negative_sampling, train_test_split_edges
from torchviz import make_dot
import neptune.new as neptune
import argparse
import traceback

from models import ShallowGNN, ToyNet
from settings import USE_SELF_EMBEDDING, DO_VISUAL, DO_REPORT_HITS, NUM_EPOCHS, URL_ERROR_CAP, NUM_MODEL_LAYERS, NUM_FEATURES, GET_DATASET_STATS, ERROR_CAP, CUR_DATASET_IN, LEARNING_RATE, WEIGHT_DECAY, USE_EARLY_STOPPING, FORCED_EPOCHS, STOPPING_CONDITION, EMBEDDING_TYPE, NUM_TESTS

parser = argparse.ArgumentParser()
parser.add_argument('--dataset', type=str, required=False)
parser.add_argument('--experiment_id', type=int, required=False)
args = parser.parse_args()

if args.dataset and not args.dataset == None:
  CUR_DATASET = args.dataset
else:
  CUR_DATASET = CUR_DATASET_IN

if args.experiment_id and not args.experiment_id == None:
  EXPERIMENT_ID = str(args.experiment_id)
else:
  EXPERIMENT_ID = None

# Helper variables
did_visual = True

class dotdict(dict):
  """dot.notation access to dictionary attributes"""
  __getattr__ = dict.get
  __setattr__ = dict.__setitem__
  __delattr__ = dict.__delitem__

def id_generator(size=6, chars=string.ascii_uppercase + string.digits):
  return ''.join(random.choice(chars) for _ in range(size))

# Main code
def main():
  # Set up Neptune
  URL_works = False
  URL_fail_count = 0
  while not URL_works:
    try:
      run = neptune.init(
        project='dylansandfelder/MScProject',
        api_token='eyJhcGlfYWRkcmVzcyI6Imh0dHBzOi8vYXBwLm5lcHR1bmUuYWkiLCJhcGlfdXJsIjoiaHR0cHM6Ly9hcHAubmVwdHVuZS5haSIsImFwaV9rZXkiOiI0NjM4OWRjMC03ZmZmLTQ2NGYtYWI5OC00YTFhNWVmNTQ0MzgifQ==',
      )
      URL_works = True
    except:
      URL_fail_count = URL_fail_count + 1
    if URL_fail_count >= URL_ERROR_CAP:
      print('Could not connect to Neptune')
      return
  if EXPERIMENT_ID:
    params = {'learning_rate': LEARNING_RATE, 'dataset': CUR_DATASET, 'num_feats': NUM_FEATURES, 'optimizer': 'Adam', 'embedding_type': EMBEDDING_TYPE, 'experiment_tag': EMBEDDING_TYPE + EXPERIMENT_ID}
  else:
    params = {'learning_rate': LEARNING_RATE, 'dataset': CUR_DATASET, 'num_feats': NUM_FEATURES, 'optimizer': 'Adam', 'embedding_type': EMBEDDING_TYPE}
  run['parameters'] = params
  run['model_source_code'].upload_files('/home/ball5842/python/LinkPrediction/models.py', wait=True)
  run['hyper_parameter_code'].upload_files('/home/ball5842/python/LinkPrediction/settings.py', wait=True)

  # Check if we are running visuals
  global did_visual
  if DO_VISUAL:
      did_visual = False

  best_test_acc_avg = 0.0
  best_val_acc_avg = 0.0
  hits_avg = None
  cur_trial = 1
  num_errors = 0
  unique_id = id_generator()

  while cur_trial <= NUM_TESTS:
    try:
      # Load device and dataset
      device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
      num_features = NUM_FEATURES
      if CUR_DATASET == 'Cora' or CUR_DATASET == 'Citeseer' or CUR_DATASET == 'Pubmed':
        dataset = Planetoid('/home/ball5842/datasets/' + CUR_DATASET, CUR_DATASET)
        data = dataset[0].to(device)
        relations = torch.zeros(data.edge_index.shape[1], 1, dtype=torch.long, device=device)
        data.edge_type = torch.flatten(relations)
        neg_relations = torch.zeros(data.edge_index.shape[1], 1, dtype=torch.int, device=device)
        num_relations = 1
        num_features = dataset.num_node_features
      elif CUR_DATASET == 'WordNet18RR':
        dataset = WordNet18RR('/home/ball5842/datasets/WordNet18RR')
        dataset.process()
        data = dataset[0].to(device)
        relations = torch.reshape(data.edge_type, (data.edge_index.shape[1], 1))
        num_relations = 11
        #neg_relations = torch.randint(0, 11, (data.edge_index.shape[1], 1), dtype=torch.int, device=device)
        if EMBEDDING_TYPE == 'ComplEx':
          word_x = torch.ones(40943, num_features*2, dtype=torch.float, device=device)
        else:
          word_x = torch.ones(40943, num_features, dtype=torch.float, device=device)
      elif CUR_DATASET == 'FB15k-237':
        dataset = RelLinkPredDataset('/home/ball5842/datasets/FB15k-237', CUR_DATASET)
        dataset.process()
        data = dataset[0].to(device)
        relations = torch.reshape(data.edge_type, (data.edge_index.shape[1], 1))
        num_relations = 474
        #neg_relations = torch.randint(0, 237, (data.edge_index.shape[1], 1), dtype=torch.int, device=device)
        if EMBEDDING_TYPE == 'ComplEx':
          word_x = torch.ones(14541, num_features*2, dtype=torch.float, device=device)
        else:
          word_x = torch.ones(14541, num_features, dtype=torch.float, device=device)
      elif len(CUR_DATASET) >= 6 and CUR_DATASET[0:6] == 'pickle':
        pick_file = open('/home/ball5842/datasets/grail/' + CUR_DATASET[6:] + '_train.pickle', 'rb')
        data = pickle.load(pick_file)
        pick_file.close()
        num_relations = data.num_relations
        if EMBEDDING_TYPE == 'ComplEx':
          word_x = torch.ones(data.num_nodes, num_features*2, dtype=torch.float, device=device)
        else:
          word_x = torch.ones(data.num_nodes, num_features, dtype=torch.float, device=device)

      # Use train_test_split_edges to create neg and positive edges
      data.train_mask = data.val_mask = data.test_mask = data.y = None
      transform = T.RandomLinkSplit(num_val=0.05, num_test=0.1, add_negative_train_samples=False, key='edge_type')
      if len(CUR_DATASET) >= 6 and CUR_DATASET[0:6] == 'pickle':
        pick_file = open('/home/ball5842/datasets/grail/' + CUR_DATASET[6:] + '_train.pickle', 'rb')
        train_pickle = pickle.load(pick_file)
        pick_file.close()
        data_train = {}
        data_train['edge_type_index'] = train_pickle.edge_index
        data_train['edge_type'] = train_pickle.edge_type + 1
        data_train['num_nodes'] = train_pickle.num_nodes
        data_train = dotdict(data_train)
        pick_file = open('/home/ball5842/datasets/grail/' + CUR_DATASET[6:] + '_valid.pickle', 'rb')
        valid_pickle = pickle.load(pick_file)
        pick_file.close()
        data_val = {}
        data_val['edge_type_index'] = valid_pickle.edge_index
        data_val['edge_type'] = valid_pickle.edge_type + 1
        data_val['num_nodes'] = valid_pickle.num_nodes
        data_val = dotdict(data_val)
        pick_file = open('/home/ball5842/datasets/grail/' + CUR_DATASET[6:] + '_test.pickle', 'rb')
        test_pickle = pickle.load(pick_file)
        pick_file.close()
        data_test = {}
        data_test['edge_type_index'] = test_pickle.edge_index
        data_test['edge_type'] = test_pickle.edge_type + 1
        data_test['num_nodes'] = test_pickle.num_nodes
        data_test = dotdict(data_test)
      elif CUR_DATASET == 'FB15k-237':
        data_train = Data(num_nodes=data.num_nodes, edge_index=data.edge_index, edge_type=data.edge_type, edge_type_index=data.edge_index)
        data_val = Data(num_nodes=data.num_nodes, edge_index=data.valid_edge_index, edge_type=data.valid_edge_type, edge_type_index=data.valid_edge_index)
        data_test = Data(num_nodes=data.num_nodes, edge_index=data.test_edge_index, edge_type=data.test_edge_type, edge_type_index=data.test_edge_index)
        #data_train, data_val, data_test = transform(data)
      elif CUR_DATASET == 'WordNet18RR':
        data_train = Data(num_nodes=data.num_nodes, edge_index=data.edge_index[:, dataset[0].train_mask], edge_type=data.edge_type[dataset[0].train_mask], edge_type_index=data.edge_index[:, dataset[0].train_mask])
        data_val = Data(num_nodes=data.num_nodes, edge_index=data.edge_index[:, dataset[0].val_mask], edge_type=data.edge_type[dataset[0].val_mask], edge_type_index=data.edge_index[:, dataset[0].val_mask])
        data_test = Data(num_nodes=data.num_nodes, edge_index=data.edge_index[:, dataset[0].test_mask], edge_type=data.edge_type[dataset[0].test_mask], edge_type_index=data.edge_index[:, dataset[0].test_mask])
        #data_train, data_val, data_test = transform(data)
      else:
        data_train = {}
        data_train['edge_type_index'] = data.train_edge_index
        data_train['edge_type'] = data.train_edge_type + 1
        data_train['num_nodes'] = data.num_nodes
        data_train = dotdict(data_train)
        data_val = {}
        data_val['edge_type_index'] = data.valid_edge_index
        data_val['edge_type'] = data.valid_edge_type + 1
        data_val['num_nodes'] = data.num_nodes
        data_val = dotdict(data_val)
        data_test = {}
        data_test['edge_type_index'] = data.test_edge_index
        data_test['edge_type'] = data.test_edge_type + 1
        data_test['num_nodes'] = data.num_nodes
        data_test = dotdict(data_test)

      neg_val_index = negative_sampling(
          edge_index=data_val.edge_type_index, # Positive edges
          num_nodes=data_val.num_nodes, # Number of nodes
          num_neg_samples=data_val.edge_type_index.size(1)) # Number of neg_sample equal to number of pos_edges

      neg_test_index = negative_sampling(
          edge_index=data_test.edge_type_index, # Positive edges
          num_nodes=data_test.num_nodes, # Number of nodes
          num_neg_samples=data_test.edge_type_index.size(1)) # Number of neg_sample equal to number of pos_edges

      neg_relations_val = torch.randint(1, num_relations + 1, data_val.edge_type.shape, dtype=torch.int, device=device)
      neg_relations_test = torch.randint(1, num_relations + 1, data_test.edge_type.shape, dtype=torch.int, device=device)

      model = ShallowGNN(num_features, num_relations+1, NUM_MODEL_LAYERS, num_features, num_features, num_relations, EMBEDDING_TYPE).to(device)
      optimizer = torch.optim.Adam(params=model.parameters(), lr=LEARNING_RATE, weight_decay=WEIGHT_DECAY)


      def get_link_labels(data, neg_edge_index):
        E = data.edge_type_index.size(1) + neg_edge_index.size(1)
        link_labels = torch.zeros(E, dtype=torch.float, device=device) - 1.
        link_labels[:data.edge_type_index.size(1)] = 1.
        return link_labels


      def train():
        model.train()

        neg_edge_index = negative_sampling(
          edge_index=data_train.edge_type_index, # Positive edges
          num_nodes=data_train.num_nodes, # Number of nodes
          num_neg_samples=data_train.edge_type_index.size(1)) # Number of neg_sample equal to number of pos_edges
        neg_relations_train = torch.randint(1, num_relations + 1, data_train.edge_type.shape, dtype=torch.int, device=device)
        optimizer.zero_grad()

        if (len(CUR_DATASET) >= 6 and CUR_DATASET[0:6] == 'pickle') or CUR_DATASET == 'WordNet18RR' or CUR_DATASET == 'FB15k-237':
          z = model.encode(word_x, data_train.edge_type_index, data_train)
        else:
          z = model.encode(data_train.x, data_train.edge_type_index, data_train)

        # Do visual if we need it
        if DO_VISUAL:
          global did_visual
          if not did_visual:
            make_dot(z, show_attrs=True, show_saved=True).render(filename='CurrentGraph' + unique_id, format='png', outfile='/home/ball5842/models/CurrentGraph' + unique_id + '.png')
            run['backward_graph'].upload_files('/home/ball5842/models/CurrentGraph' + unique_id + '.png')
            did_visual = True

        #link_logits = model.decode(z, data_train.edge_type_index, neg_edge_index, data_train.edge_type, neg_relations_train, num_relations + 1)
        link_logits = model.decode(z, data_train.edge_type_index, neg_edge_index, data_train.edge_type, data_train.edge_type, num_relations + 1)
        link_labels = get_link_labels(data_train, neg_edge_index)
        x1 = torch.flatten(link_logits)[0:data_train.edge_type_index.shape[1]]
        x1 = torch.reshape(x1, (-1, 1))
        x2 = torch.flatten(link_logits)[data_train.edge_type_index.shape[1]:]
        x2 = torch.reshape(x2, (-1, 1))
        total_x = torch.cat([x1, x2], dim=-1)
        flips = torch.randint(0,  2, (x1.shape[0], 1))
        #loss = F.margin_ranking_loss(torch.flatten(total_x.gather(1, flips)), torch.flatten(total_x.gather(1, ((flips * -1) + 1))), torch.flatten((flips * -2) + 1))
        new_x1 = torch.flatten(total_x.gather(1, flips))
        new_x2 = torch.flatten(total_x.gather(1, ((flips * -1) + 1)))
        new_target = torch.flatten((flips * -2) + 1)
        loss_fn = torch.nn.MarginRankingLoss(margin=0.1)
        loss = loss_fn(new_x1, new_x2, new_target)
        loss.backward()
        optimizer.step()
        return loss


      @torch.no_grad()
      def test():
        model.eval()
        perfs = []
        for prefix in ['val', 'test']:
          if prefix == 'val':
            pos_edge_index = data_val.edge_type_index
            neg_edge_index = neg_val_index
            cur_data = data_val
            neg_rel = neg_relations_val
          else:
            pos_edge_index = data_test.edge_type_index
            neg_edge_index = neg_test_index
            cur_data = data_test
            neg_rel = neg_relations_test

          if (len(CUR_DATASET) >= 6 and CUR_DATASET[0:6] == 'pickle') or CUR_DATASET == 'WordNet18RR' or CUR_DATASET == 'FB15k-237':
            z = model.encode(word_x, data_train.edge_type_index, data_train)
          else:
            z = model.encode(data_train.x, data_train.edge_type_index, data_train)
          link_logits = model.decode(z, pos_edge_index, neg_edge_index, cur_data.edge_type, neg_rel, num_relations + 1) # decode test or val
          link_probs = link_logits.sigmoid() # apply sigmoid

          link_labels = get_link_labels(cur_data, neg_edge_index) # get link

          perfs.append(average_precision_score(link_labels.cpu(), link_probs.cpu())) # compute auc pr score
        return perfs

      @torch.no_grad()
      def hits_transductive(n_list):
        model.eval()
        train_size = data_train.edge_type_index.size(1)
        val_size = data_val.edge_type_index.size(1)
        test_size = data_test.edge_type_index.size(1)
        total_size = train_size + val_size + test_size
        train_dict = {}
        for i in range(train_size):
          train_dict[str(int(data_train.edge_type_index[0][i])) + ',' + str(int(data_train.edge_type_index[1][i])) + 'r' + str(int(data_train.edge_type[i]))] = True
        val_dict = {}
        for i in range(val_size):
          val_dict[str(int(data_val.edge_type_index[0][i])) + ',' + str(int(data_val.edge_type_index[1][i])) + 'r' + str(int(data_val.edge_type[i]))] = True

        perfs = []
        if (len(CUR_DATASET) >= 6 and CUR_DATASET[0:6] == 'pickle') or CUR_DATASET == 'WordNet18RR' or CUR_DATASET == 'FB15k-237':
          z = model.encode(word_x, data_train.edge_type_index, data_train)
        else:
          z = model.encode(data_train.x, data_train.edge_type_index, data_train)

        ranks = []

        for i in range(test_size):
          if i % 1000 == 0:
            print(str(i) + '/' + str(test_size) + ':')
          head = int(data_test.edge_type_index[0][i])
          tail = int(data_test.edge_type_index[1][i])
          rel = int(data_test.edge_type[i])
          neg_test_index_hits1 = []
          neg_test_index_hits2 = []
          for x in range(data_test.num_nodes):
            pot_key = str(head) + ',' + str(tail) + 'r' + str(rel)
            if not pot_key in train_dict and not pot_key in val_dict:
              if not x == tail:
                neg_test_index_hits1.append([head, x])
              if not x == head:
                neg_test_index_hits2.append([x, tail])

          if not len(neg_test_index_hits1) == 0:
            neg_test_index_hits1 = torch.transpose(torch.tensor(neg_test_index_hits1), 0, 1)
            neg_relations_test_hits = torch.ones((neg_test_index_hits1.size(1),), dtype=torch.int, device=device) * rel
            link_logits = model.decode(z, data_test.edge_type_index, neg_test_index_hits1, data_test.edge_type, neg_relations_test_hits, num_relations + 1) # decode test or val
            link_probs = link_logits.sigmoid().tolist() # apply sigmoid
            scores = [link_probs[i]] + link_probs[test_size:]
            rank = neg_test_index_hits1.size(1) - rankdata(scores, method='min')[0] + 2.0
            ranks.append(float(rank))
            if i % 1000 == 0:
              print('All Neg Tails: ' + str(int(neg_test_index_hits1.size(1))) + ', Scores length: ' +  str(len(scores)))
              print('Raw Rank: ' + str(rankdata(scores, method='min')[0]) + ', Rank: ' +  str(rank))
          else:
            rank = 1.0
            ranks.append(float(rank))

          if not len(neg_test_index_hits2) == 0:
            neg_test_index_hits2 = torch.transpose(torch.tensor(neg_test_index_hits2), 0, 1)
            neg_relations_test_hits = torch.ones((neg_test_index_hits2.size(1),), dtype=torch.int, device=device) * rel
            link_logits = model.decode(z, data_test.edge_type_index, neg_test_index_hits2, data_test.edge_type, neg_relations_test_hits, num_relations + 1) # decode test or val
            link_probs = link_logits.sigmoid().tolist() # apply sigmoid
            scores = [link_probs[i]] + link_probs[test_size:]
            rank = neg_test_index_hits2.size(1) - rankdata(scores, method='min')[0] + 2.0
            ranks.append(float(rank))
            if i % 1000 == 0:
              print('All Neg Heads: ' + str(int(neg_test_index_hits2.size(1))) + ', Scores length: ' +  str(len(scores)))
              print('Raw Rank: ' + str(rankdata(scores, method='min')[0]) + ', Rank: ' +  str(rank))
          else:
            rank = 1.0
            ranks.append(float(rank))

        hits_list = []
        print('Mean Rank: ' + str(np.mean(np.array(ranks))))
        mrr = 1.0 / np.array(ranks)
        mrr = np.mean(mrr)
        hits_list.append(mrr)
        for n in n_list:
          isHitList = [x for x in ranks if x <= n]
          hits_list.append(len(isHitList) / len(ranks))

        return hits_list

      @torch.no_grad()
      def hits(n_list, neg_amt):
        model.eval()
        test_size = data_test.edge_type_index.size(1)
        perfs = []
        neg_test_index_hits = negative_sampling(
          edge_index=data_test.edge_type_index, # Positive edges
          num_nodes=data_test.num_nodes, # Number of nodes
          num_neg_samples=test_size * neg_amt) # Number of neg_sample equal to number of pos_edges
        neg_relations_test_hits = torch.randint(1, num_relations + 1, (data_test.edge_type.size(0) * neg_amt,), dtype=torch.int, device=device)
        if (len(CUR_DATASET) >= 6 and CUR_DATASET[0:6] == 'pickle') or CUR_DATASET == 'WordNet18RR' or CUR_DATASET == 'FB15k-237':
          z = model.encode(word_x, data_train.edge_type_index, data_train)
        else:
          z = model.encode(data_train.x, data_train.edge_type_index, data_train)
        link_logits = model.decode(z, data_test.edge_type_index, neg_test_index_hits, data_test.edge_type, neg_relations_test_hits, num_relations + 1) # decode test or val
        link_probs = link_logits.sigmoid().tolist() # apply sigmoid

        ranks = []
        for i in range(test_size):
          scores = [link_probs[i]] + link_probs[test_size + (i * neg_amt):test_size + ((i+1) * neg_amt)]
          rank = neg_amt - rankdata(scores, method='min')[0] + 2.0
          ranks.append(rank)

        hits_list = []
        mrr = np.mean(1.0 / np.array(ranks))
        hits_list.append(mrr)
        for n in n_list:
          isHitList = [x for x in ranks if x <= n]
          hits_list.append(len(isHitList) / len(ranks))

        return hits_list


      best_val_perf = test_perf = 0

      # Primary training loop
      if USE_EARLY_STOPPING:
        stop_training = False
        stopping_counter = 0
        cur_epoch = 1
        while not stop_training:
          train_loss = train()
          run['training_loss'].log(train_loss)
          val_perf, tmp_test_perf = test()
          run['val_roc_auc'].log(val_perf)
          run['test_roc_auc'].log(tmp_test_perf)
          if USE_SELF_EMBEDDING:
            run['self_loop_embedding_norm'].log(float(torch.linalg.norm(model.get_self_embedding())))
          if val_perf > best_val_perf:
            best_val_perf = val_perf
            test_perf = tmp_test_perf
            stopping_counter = 0
            model_f = open('/home/ball5842/models/cur_model' + unique_id + '.p', 'wb')
            pickle.dump(model, model_f)
            model_f.close()
            if DO_REPORT_HITS:
              if CUR_DATASET == 'WordNet18RR' or CUR_DATASET == 'FB15k-237':
                hit_nums = hits_transductive([1, 3, 5, 10])
              else:
                hit_nums = hits([1, 3, 5, 10], 50)
              print('MRR: ' + str(hit_nums[0]))
              print('Hits@1: ' + str(hit_nums[1]))
              print('Hits@3: ' + str(hit_nums[2]))
              print('Hits@5: ' + str(hit_nums[3]))
              print('Hits@10: ' + str(hit_nums[4]))
          else:
            stopping_counter = stopping_counter + 1
          run['cur_best_test'].log(test_perf)
          run['cur_best_val'].log(best_val_perf)
          log = 'Epoch: {:03d}, Loss: {:.4f}, Val: {:.4f}, Test: {:.4f}'
          if cur_epoch % 10 == 0:
            print(log.format(cur_epoch, train_loss, best_val_perf, test_perf))
          if cur_epoch <= FORCED_EPOCHS:
            stopping_counter = 0
          else:
            if stopping_counter >= STOPPING_CONDITION:
              stop_training = True
              print(log.format(cur_epoch, train_loss, best_val_perf, test_perf))
              model_f = open('/home/ball5842/models/cur_model' + unique_id + '.p', 'rb')
              model = pickle.load(model_f)
              model_f.close()
          cur_epoch = cur_epoch + 1
      else:
        for epoch in range(1, NUM_EPOCHS + 1):
          train_loss = train()
          run['training_loss'].log(train_loss)
          val_perf, tmp_test_perf = test()
          run['val_roc_auc'].log(val_perf)
          run['test_roc_auc'].log(tmp_test_perf)
          if val_perf > best_val_perf:
            best_val_perf = val_perf
            test_perf = tmp_test_perf
          log = 'Epoch: {:03d}, Loss: {:.4f}, Val: {:.4f}, Test: {:.4f}'
          if epoch % 10 == 0:
            print(log.format(epoch, train_loss, best_val_perf, test_perf))
        model_f = open('/home/ball5842/models/cur_model' + unique_id + '.p', 'wb')
        pickle.dump(model, model_f)
        model_f.close()

      if (len(CUR_DATASET) >= 6 and CUR_DATASET[0:6] == 'pickle') or CUR_DATASET == 'WordNet18RR' or CUR_DATASET == 'FB15k-237':
        z = model.encode(word_x, data_train.edge_type_index, data_train)
      else:
        z = model.encode(data.x, data_train.edge_type_index, data_train)
      final_edge_index = model.decode_all(z)

      if CUR_DATASET == 'WordNet18RR' or CUR_DATASET == 'FB15k-237':
        hit_nums = hits_transductive([1, 3, 5, 10])
      else:
        hit_nums = hits([1, 3, 5, 10], 50)
      print('MRR: ' + str(hit_nums[0]))
      print('Hits@1: ' + str(hit_nums[1]))
      print('Hits@3: ' + str(hit_nums[2]))
      print('Hits@5: ' + str(hit_nums[3]))
      print('Hits@10: ' + str(hit_nums[4]))

      run['model_file'].upload_files('/home/ball5842/models/cur_model' + unique_id + '.p', wait=True)

      if GET_DATASET_STATS and cur_trial == 1:
        stat_num_nodes_train = data_train.num_nodes
        stat_num_nodes_val = data_val.num_nodes
        stat_num_nodes_test = data_test.num_nodes
        stat_num_edges_train = int(data_train.edge_type_index.shape[1])
        stat_num_edges_val = int(data_val.edge_type_index.shape[1])
        stat_num_edges_test = int(data_test.edge_type_index.shape[1])
        print('Number of Nodes (Train): ' + str(stat_num_nodes_train))
        print('Number of Nodes (Val): ' + str(stat_num_nodes_val))
        print('Number of Nodes (Test): ' + str(stat_num_nodes_test))
        print('Number of Positive Edges (Train): ' + str(stat_num_edges_train))
        print('Number of Positive Edges (Val): ' + str(stat_num_edges_val))
        print('Number of Positive Edges (Test): ' + str(stat_num_edges_test))
        print('Number of relations: ' + str(num_relations))

      if DO_VISUAL and cur_trial == 1:
        os.remove('/home/ball5842/models/CurrentGraph' + unique_id + '.png')
        os.remove('/home/ball5842/python/CurrentGraph' + unique_id)
      os.remove('/home/ball5842/models/cur_model' + unique_id + '.p')

    except Exception as e:
      # In case we actually want to see the errors:
      raise e
      traceback.print_exc()
      print(e)
      num_errors = num_errors + 1
      if num_errors > ERROR_CAP:
        print('We reached the error cap of: ' + str(ERROR_CAP))
        return
      else:
        continue

    best_test_acc_avg = best_test_acc_avg + test_perf
    best_val_acc_avg = best_val_acc_avg + best_val_perf
    if hits_avg:
      hits_avg[0] = hits_avg[0] + hit_nums[0]
      hits_avg[1] = hits_avg[1] + hit_nums[1]
      hits_avg[2] = hits_avg[2] + hit_nums[2]
      hits_avg[3] = hits_avg[3] + hit_nums[3]
      hits_avg[4] = hits_avg[4] + hit_nums[4]
    else:
      hits_avg = hit_nums[:]
    cur_trial = cur_trial + 1

  print('Final experimental results:')
  best_test_acc_avg = best_test_acc_avg / NUM_TESTS
  best_val_acc_avg = best_val_acc_avg / NUM_TESTS
  hits_avg[0] = hits_avg[0] / NUM_TESTS
  hits_avg[1] = hits_avg[1] / NUM_TESTS
  hits_avg[2] = hits_avg[2] / NUM_TESTS
  hits_avg[3] = hits_avg[3] / NUM_TESTS
  hits_avg[4] = hits_avg[4] / NUM_TESTS
  print('MRR: ' + str(hits_avg[0]))
  print('Hits@1: ' + str(hits_avg[1]))
  print('Hits@3: ' + str(hits_avg[2]))
  print('Hits@5: ' + str(hits_avg[3]))
  print('Hits@10: ' + str(hits_avg[4]))
  run['best_test'].log(best_test_acc_avg)
  run['best_val'].log(best_val_acc_avg)
  run['mrr_avg'].log(hits_avg[0])
  run['hits1'].log(hits_avg[1])
  run['hits3'].log(hits_avg[2])
  run['hits5'].log(hits_avg[3])
  run['hits10'].log(hits_avg[4])
  run['num_tests'].log(NUM_TESTS)
  run.stop()

if __name__ == "__main__":
  main()

